#*************************************************************************
#                            client_lib.tcl
#                            ==============
#                           Uwe Berger; 2007
#
# Ein paar Standard-Routinen zur Erstellung eines Ansteuer-Clients fuer
# cube_viewer.tcl
#
# Wie muss ein Client aussehen:
# -----------------------------
#
# client_lib.tcl muss via source-Anweisung eingezogen werden. Dabei werden
# einige Kommunikationsroutinen deklariert und ausgefhrt, die die TCP-Ver-
# bindung zum Server herstellen. Desweiteren wird die Dimension des Cubes
# vom Server abgefragt (proc get_xyz) und in der globalen Variable gvar ab-
# gelegt (gvar(x), gvar(y), gvar(z)). Am Besten erfolgt die Einbindung 
# dieser Lib gleich als erste Anweisung! IP-Adresse und -Port werden in den
# beiden Variablen gvar(host) und gvar(port) in dieser Lib definiert.
#
# Im Client muss zwingend die Procedure"display_led" definiert sein (ist
# in client_lib.tcl -> "proc update_ok {}..." so verdrahtet)
#
# Hat die globale Variable gvar(update_refresh) einen Wert groesser
# Null, wird "display_led" zyklisch aufgerufen (Zyklusrate entspricht
# dem Wert von gvar(update_refresh) -> 1 = 1ms). In client_lib.tcl
# wird dieser Wert mit 0 initiiert, kann aber im Client neu gesetzt
# werden.
#
# Die Uebertragung des anzuzeigenden Zustandes der LEDs (steht in gvar(led))
# erfolgt mit dem Senden des Server-Befehles "set_led" an den Selbigen. 
# Das Senden des LED-Zustandes ist in der Routine "send_led" gekapselt.
#
# Der Server antwortet nach dem Verarbeiten mit dem Befehl "update_ok".
# Erst nach dem Empfang dieser Zeichenfolge wird, wenn so konfiguriert,
# die Routine "display_led" erneut aufgerufen, um z.B. das naechste
# Bild zu berechnen o.�.
#
# Saemtliche Zaehler/Zustaende usw., die einem erneuten Aufruf von
# "display_led" uebergeben werden sollen, sollten in globalen 
# Variablen (z.B. gvar()) abgelegt werden.
#
# "display_led" muss einmal im Client-Programm aufgerufen werden, er-
# neute Aufrufe erfolgen ggf. (gvar(update_refresh) > 0) automatisch.
#
#*************************************************************************

# IP-Adresse des Host auf dem cube_viewer.tcl laeuft
set gvar(host) "etherrape"  

# Port-Adresse auf der cube_viewer.tcl lauscht
set gvar(port) 12345        

# initial keine LEDs (wird spaeter durch get_xyz ermittelt)
set gvar(led_count)      0  

# initial keine Refresh-Zyklen von display_led
set gvar(update_refresh) 0  

#*************************************************************************
# Routinen
#
#
#----- receice -----
#
# Wird aufgerufen, wenn etwas im Empfangspuffer drin ist...
#
proc receive {sock} {
	if {[eof $sock] || [catch {gets $sock line}]} {
		close $sock
	} else {
		if {$line != ""} {
			eval $line
		}
	}
}


#----- update_ok -----
#   (Serverbefehl)
#
proc update_ok {} {
	global gvar
	if {$gvar(update_refresh) > 0} {after $gvar(update_refresh) [list display_led]}
}


#----- set_xyz -----
#   (Serverbefehl)
#
proc set_xyz {x y z} {
	global gvar
	set gvar(x) $x
	set gvar(y) $y
	set gvar(z) $z
	set gvar(led_count) [expr $gvar(x) * $gvar(y) * $gvar(z)]
	init_led
}


#----- init_led -----
#
proc init_led {} {
	global gvar
	set gvar(led) {}
	for {set i 0} {$i < $gvar(led_count)} {incr i} {set gvar(led) [linsert $gvar(led) end 0]}
}

proc send_led {} {
	global gvar
	send_cmd [list set_led [join $gvar(led) ""]]
}


#----- send_cmd -----
#
proc send_cmd {cmd} {
	global gvar
        if { [ catch {puts $gvar(sock) $cmd; flush $gvar(sock)} ] } {
          puts stderr "No Socket ! No Cube !"
        }
}


#---- draw_point ----
#
proc draw_point {x y z} {
	global gvar
	# nur Punkte zeichnen, die im Bereich liegen, alle anderen werden uebergangen...
	if {(($x >= 0) && ($x < $gvar(x))) && (($y >= 0) && ($y < $gvar(y))) && (($z >= 0) && ($z < $gvar(z)))} {
		set i [expr $x + ($y * $gvar(x)) + ($z * $gvar(x) * $gvar(y))]
		set gvar(led) [lreplace $gvar(led) $i $i 1]
	}
}


#---- get_distance ----
#
proc get_distance {x1 y1 z1 x2 y2 z2} {
	set dist [expr sqrt([expr pow([expr $x2 - $x1], 2)\
                                + pow([expr $y2 - $y1], 2)\
			        + pow([expr $z2 - $z1], 2)])]
	return $dist
}


#---- draw_line ----
#
proc draw_line {x1 y1 z1 x2 y2 z2} {
	draw_point $x1 $y1 $z1
	set d_x [expr $x2 - $x1]
	set d_y [expr $y2 - $y1]
	set d_z [expr $z2 - $z1]
	set dx [expr abs($d_x)]
	set dy [expr abs($d_y)]
	set dz [expr abs($d_z)]
	set dd 0
	if {($dx > 0) && ($dx >= $dy) && ($dx >= $dz)} {
		set dd $dx
	} else {
		if {($dy > 0) && ($dy >= $dx) && ($dy >= $dz)} {
			set dd $dy
		} else {
			set dd $dz
		}
	}
	set i 1
	while {$i <= $dd} {
		set x [expr $x1 + round($d_x * $i / $dd)]
		set y [expr $y1 + round($d_y * $i / $dd)]
		set z [expr $z1 + round($d_z * $i / $dd)]
		draw_point $x $y $z
		incr i
	}
}


#****************** Initialisierung ******************************
#
# TCP/IP-Socket zum Server oeffnen
proc connection_start {} {
  global gvar

  if { [ catch {socket $gvar(host) $gvar(port)} gvar(sock) ] } {
    puts stderr "No Socket ! No Cube !"
  } else {
    fconfigure $gvar(sock) -buffering line
    fileevent  $gvar(sock) readable [list receive $gvar(sock)]

    # Dimmension des Cubes vom Server holen
    send_cmd get_xyz 

    # warten, bis Replay vom Server kam und Ergebnis verarbeitet wurde
    vwait gvar(led_count)
  }
}

connection_start
